@@ -1,13 +1,15 @@
-import argparse
-import logging
 import sys
 import json
+import shutil
 import typing
+import logging
 import pathlib
+import argparse
 import itertools
 import collections
 import collections.abc
 import urllib.parse
+import urllib.request
 
 import attr
 import csvw
@@ -19,19 +21,23 @@ from clldutils.misc import log_or_raise, lazyproperty
 from clldutils import jsonlib
 
 from pycldf.sources import Sources
-from pycldf.util import pkg_path, resolve_slices, DictTuple
+from pycldf.util import pkg_path, resolve_slices, DictTuple, sanitize_url
 from pycldf.terms import term_uri, Terms, TERMS, get_column_names, URL as TERMS_URL
 from pycldf.validators import VALIDATORS
 from pycldf import orm
 
 __all__ = [
     'Dataset', 'Generic', 'Wordlist', 'ParallelText', 'Dictionary', 'StructureDataset',
-    'iter_datasets', 'sniff']
+    'iter_datasets', 'sniff', 'SchemaError']
 
 MD_SUFFIX = '-metadata.json'
 ORM_CLASSES = {cls.component_name(): cls for cls in orm.Object.__subclasses__()}
 
 
+class SchemaError(KeyError):
+    pass
+
+
 @attr.s
 class Module(object):
     """
@@ -96,7 +102,10 @@ def make_column(spec):
 
 class GitRepository(object):
     def __init__(self, url, clone=None, version=None, **dc):
-        self.url = url
+        # We remove credentials from the URL immediately to make sure this isn't leaked into
+        # CLDF metadata. Such credentials might be present in URLs read via gitpython from
+        # remotes.
+        self.url = sanitize_url(url)
         self.clone = clone
         self.version = version
         self.dc = dc
@@ -347,7 +356,7 @@ class Dataset(object):
           propertyUrl of a column
         - the name of a column
 
-        :raises KeyError: If no matching table or column is found.
+        :raises SchemaError: If no matching table or column is found.
         """
         if isinstance(item, tuple):
             table, column = item
@@ -364,19 +373,36 @@ class Dataset(object):
                         or t.url.string == table:
                     break
             else:
-                raise KeyError(table)
+                raise SchemaError('Dataset has no table "{}"'.format(table))
         else:
-            t = table
+            if any(table is tt for tt in self.tables):
+                t = table
+            else:
+                raise SchemaError('Dataset has no table "{}"'.format(table))
 
         if not column:
             return t
 
+        if isinstance(column, Column):
+            if any(column is c for c in t.tableSchema.columns):
+                return column
+            else:
+                raise SchemaError('Dataset has no column "{}" in table "{}"'.format(
+                    column.name, t.url))
+
         uri = term_uri(column, terms=TERMS.by_uri)
         for c in t.tableSchema.columns:
             if (c.propertyUrl and c.propertyUrl.uri == uri) or c.header == column:
                 return c
 
-        raise KeyError(column)
+        raise SchemaError('Dataset has no column "{}" in table "{}"'.format(column, t.url))
+
+    def __delitem__(self, key):
+        thing = self[key]
+        if isinstance(thing, Column):
+            self.remove_columns(self[key[0]], thing)
+        else:
+            self.remove_table(thing)
 
     def __contains__(self, item) -> bool:
         """
@@ -394,7 +420,7 @@ class Dataset(object):
         """
         try:
             return self[item]
-        except KeyError:
+        except SchemaError:
             return default
 
     def get_foreign_key_reference(self, table, column) \
@@ -522,7 +548,6 @@ class Dataset(object):
             if other_table.url == component.url:
                 raise ValueError('tables must have distinct url properties')
 
-        self.add_columns(component, *cols)
         try:
             table_type = self.get_tabletype(component)
         except ValueError:
@@ -535,8 +560,9 @@ class Dataset(object):
                     continue
                 if other_table_type == table_type:
                     raise ValueError('components must not be added twice')
-
         self.tables.append(component)
+        self.add_columns(component, *cols)
+
         component._parent = self.tablegroup
         self.auto_constraints(component)
         return component
@@ -674,7 +700,7 @@ class Dataset(object):
     #
     # Methods to read data
     #
-    def iter_rows(self, table, *cols) -> typing.Iterable[dict]:
+    def iter_rows(self, table, *cols) -> typing.Iterator[dict]:
         """
         Iterate rows in a table, resolving CLDF property names to local column names.
 
@@ -776,7 +802,7 @@ class Dataset(object):
         """
         return self.sources.write(self.bibpath)
 
-    def write(self, fname=None, **table_items: typing.Dict[str, typing.List[dict]]):
+    def write(self, fname=None, **table_items: typing.List[dict]):
         """
         Write metadata, sources and data. Metadata will be written to `fname` (as interpreted in
         :meth:`pycldf.dataset.Dataset.write_metadata`); data files will be written to the file
@@ -793,6 +819,55 @@ class Dataset(object):
             table.common_props['dc:extent'] = table.write(items)
         self.write_metadata(fname)
 
+    def copy(self, dest: typing.Union[str, pathlib.Path], mdname: str = None) -> pathlib.Path:
+        """
+        Copy metadata, data and sources to files in `dest`.
+
+        :param dest: Destination directory.
+        :param mdname: Name of the new metadata file.
+        :return: Path of the new CLDF metadata file.
+
+        This can be used together with :func:`iter_datasets` to extract CLDF data from their
+        curation context, e.g. cldfbench-curated datasets from the repository they are curated in.
+
+        .. code-block:: python
+
+            >>> from pycldf import iter_datasets
+            >>> for ds in iter_datasets('tests/data'):
+            ...     if 'with_examples' in ds.directory.name:
+            ...         ds.copy('some_directory', mdname='md.json')
+        """
+        dest = pathlib.Path(dest)
+        if not dest.exists():
+            dest.mkdir(parents=True)
+
+        from_url = is_url(self.tablegroup.base)
+        ds = Dataset.from_metadata(self.tablegroup.base if from_url else self.tablegroup._fname)
+
+        _getter = urllib.request.urlretrieve if from_url else shutil.copy
+        try:
+            _getter(self.bibpath, dest / self.bibname)
+            ds.properties['dc:source'] = self.bibname
+        except:  # pragma: no cover # noqa
+            # Sources are optional
+            pass
+
+        for table in ds.tables:
+            fname = table.url.resolve(table.base)
+            name = pathlib.Path(urllib.parse.urlparse(fname).path).name if from_url else fname.name
+            _getter(fname, dest / name)
+            table.url = Link(name)
+
+            for fk in table.tableSchema.foreignKeys:
+                fk.reference.resource = Link(pathlib.Path(fk.reference.resource.string).name)
+        mdpath = dest.joinpath(
+            mdname or  # noqa: W504
+            (self.tablegroup.base.split('/')[-1] if from_url else self.tablegroup._fname.name))
+        if from_url:
+            del ds.tablegroup.at_props['base']  # pragma: no cover
+        ds.write_metadata(fname=mdpath)
+        return mdpath
+
     #
     # Reporting
     #
@@ -1030,7 +1105,7 @@ def sniff(p: pathlib.Path) -> bool:
     return d.get('dc:conformsTo', '').startswith(TERMS_URL)
 
 
-def iter_datasets(d: pathlib.Path) -> typing.Iterable[Dataset]:
+def iter_datasets(d: pathlib.Path) -> typing.Iterator[Dataset]:
     """
     Discover CLDF datasets - by identifying metadata files - in a directory.
 
